@@ -57,6 +57,11 @@ static inline void __list_splice(struct list_head *list,
 	last->next = at;
 	at->prev = last;
 }
+static inline void list_splice(struct list_head *list, struct list_head *head)
+{
+	if (!list_empty(list))
+		__list_splice(list, head);
+}
 static inline void list_splice_init(struct list_head *list,
 				    struct list_head *head)
 {
@@ -73,8 +78,10 @@ extern int disable_bd_claim;
 extern int major_nr;
 extern int use_nbd_major;
 
+#ifndef NO_MORE_DEV_FS
 // use_nbd_major ? "nbd" : "drbd";
 extern char* drbd_devfs_name;
+#endif
 
 #include <linux/major.h>
 #ifdef DRBD_MAJOR
@@ -212,11 +219,11 @@ typedef struct Drbd_Conf drbd_dev;
 
 #ifdef DBG_ASSERTS
 extern void drbd_assert_breakpoint(drbd_dev*, char *, char *, int );
-# define D_ASSERT(exp)  if (!(exp)) \
-	 drbd_assert_breakpoint(mdev,#exp,__FILE__,__LINE__)
+# define D_ASSERT(exp)  do { if (!(exp)) \
+	 drbd_assert_breakpoint(mdev,#exp,__FILE__,__LINE__); } while (0)
 #else
-# define D_ASSERT(exp)  if (!(exp)) \
-	 ERR("ASSERT( " #exp " ) in %s:%d\n", __FILE__,__LINE__)
+# define D_ASSERT(exp)  do { if (!(exp)) \
+	 ERR("ASSERT( " #exp " ) in %s:%d\n", __FILE__,__LINE__); } while (0) 
 #endif
 #define ERR_IF(exp) if (({ \
 	int _b = (exp)!=0; \
@@ -611,6 +618,8 @@ struct drbd_request {
 	int rq_status;
 	struct drbd_barrier *barrier; // The next barrier.
 	drbd_bio_t *master_bio;       // master bio pointer
+	unsigned int size;
+	sector_t sector;
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 	drbd_bio_t private_bio;       // private bio struct
 #else
@@ -681,7 +690,8 @@ enum {
 	MD_DIRTY,		// current gen counts and flags not yet on disk
 	SYNC_STARTED,		// Needed to agree on the exact point in time..
 	USE_DEGR_WFC_T,		// Use degr-wfc-timeout instad of wfc-timeout.
-	CRASHED_PRIMARY         // This node was a crashed primary
+	CRASHED_PRIMARY,        // This node was a crashed primary
+	IO_FROZEN               // IO Frozen.
 };
 
 struct drbd_bitmap; // opaque for Drbd_Conf
@@ -764,7 +774,6 @@ struct Drbd_Conf {
 	atomic_t unacked_cnt;    // Need to send replys for
 	atomic_t local_cnt;      // Waiting for local disk to signal completion
 	spinlock_t req_lock;
-	spinlock_t tl_lock;
 	struct drbd_barrier* newest_barrier;
 	struct drbd_barrier* oldest_barrier;
 	unsigned long flags;
@@ -793,6 +802,8 @@ struct Drbd_Conf {
 	struct list_head done_ee;   // send ack
 	struct list_head read_ee;   // IO in progress
 	struct list_head net_ee;    // zero-copy network send in progress
+	drbd_bio_t *first_frozen_bio;
+	drbd_bio_t *last_frozen_bio;
 	spinlock_t pr_lock;
 	struct list_head app_reads;
 	struct list_head resync_reads;
@@ -824,7 +835,8 @@ extern void drbd_free_resources(drbd_dev *mdev);
 extern void tl_release(drbd_dev *mdev,unsigned int barrier_nr,
 		       unsigned int set_size);
 extern void tl_clear(drbd_dev *mdev);
-extern int tl_dependence(drbd_dev *mdev, drbd_request_t * item);
+extern void tl_resend(drbd_dev *mdev);
+extern int tl_dependence(drbd_dev *mdev, drbd_request_t * item, int free_it);
 extern void drbd_free_sock(drbd_dev *mdev);
 extern int drbd_send(drbd_dev *mdev, struct socket *sock,
 		     void* buf, size_t size, unsigned msg_flags);
@@ -855,6 +867,9 @@ extern void drbd_free_ll_dev(drbd_dev *mdev);
 extern int drbd_io_error(drbd_dev* mdev);
 extern void drbd_mdev_cleanup(drbd_dev *mdev);
 
+extern int drbd_resend_barrier(drbd_dev *mdev,struct drbd_barrier *b);
+extern int drbd_resend_dblock(drbd_dev *mdev, drbd_request_t *req);
+
 // drbd_meta-data.c (still in drbd_main.c)
 extern void drbd_md_write(drbd_dev *mdev);
 extern int drbd_md_read(drbd_dev *mdev);
@@ -1013,13 +1028,13 @@ extern void drbd_bm_unlock    (drbd_dev *mdev);
 // drbd_main.c
 extern drbd_dev *drbd_conf;
 extern int minor_count;
-extern kmem_cache_t *drbd_request_cache;
-extern kmem_cache_t *drbd_ee_cache;
+extern drbd_kmem_cache_t *drbd_request_cache;
+extern drbd_kmem_cache_t *drbd_ee_cache;
 extern mempool_t *drbd_request_mempool;
 
 // drbd_req
-#define ERF_NOTLD    2   /* do not call tl_dependence */
-extern void drbd_end_req(drbd_request_t *, int, int, sector_t);
+extern void _drbd_end_req(drbd_request_t *, int, int);
+extern void drbd_end_req(drbd_request_t *, int, int);
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 extern int drbd_make_request_24(request_queue_t *q, int rw, struct buffer_head *bio);
 #else
@@ -1027,6 +1042,7 @@ extern int drbd_make_request_26(request_queue_t *q, struct bio *bio);
 #endif
 extern int drbd_read_remote(drbd_dev *mdev, drbd_request_t *req);
 
+extern void drbd_thaw_frozen_reqs(drbd_dev *mdev);
 // drbd_fs.c
 extern char* ppsize(char* buf, size_t size);
 extern int drbd_determin_dev_size(drbd_dev*);
@@ -1316,76 +1332,36 @@ static inline void inc_ap_pending(drbd_dev* mdev)
 		    __func__ , __LINE__ ,			\
 		    atomic_read(&mdev->which))
 
-#define dec_ap_pending(mdev)					\
+#define dec_ap_pending(mdev) do {				\
 	typecheck(drbd_dev*,mdev);				\
 	if(atomic_dec_and_test(&mdev->ap_pending_cnt))		\
 		wake_up(&mdev->cstate_wait);			\
-	ERR_IF_CNT_IS_NEGATIVE(ap_pending_cnt)
+	ERR_IF_CNT_IS_NEGATIVE(ap_pending_cnt); } while (0)
 
 static inline void inc_rs_pending(drbd_dev* mdev)
 {
 	atomic_inc(&mdev->rs_pending_cnt);
 }
 
-#define dec_rs_pending(mdev)					\
+#define dec_rs_pending(mdev) do {				\
 	typecheck(drbd_dev*,mdev);				\
 	atomic_dec(&mdev->rs_pending_cnt);			\
-	ERR_IF_CNT_IS_NEGATIVE(rs_pending_cnt)
+	ERR_IF_CNT_IS_NEGATIVE(rs_pending_cnt); } while (0)
 
 static inline void inc_unacked(drbd_dev* mdev)
 {
 	atomic_inc(&mdev->unacked_cnt);
 }
 
-#if 0 && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
-/*
- * idea was to forcefully push the tcp stack whenever the
- * currently last pending packet is in the buffer.
- * should be benchmarked on some real box to see if it has any
- * effect on overall latency.
- */
-
-/* this only works with 2.6 kernels because of some conflicting defines
- * in header files included from net.tcp.h.
- */
-
-#include <net/tcp.h>
-static inline void drbd_push_msock(drbd_dev* mdev)
-{
-	struct sock    *sk;
-	struct tcp_opt *tp;
-	if (mdev->meta.socket == NULL) return;
-	sk = mdev->meta.socket->sk;
-	tp = tcp_sk(sk);
-	lock_sock(sk);
-	__tcp_push_pending_frames(sk, tp, tcp_current_mss(sk, 1), TCP_NAGLE_PUSH);
-	release_sock(sk);
-}
-
-#define dec_unacked(mdev)					\
-	might_sleep();						\
-	typecheck(drbd_dev*,mdev);				\
-	if (atomic_dec_and_test(&mdev->unacked_cnt))		\
-		drbd_push_msock(mdev);				\
-	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt);
-
-#define sub_unacked(mdev, n)					\
-	might_sleep();						\
-	typecheck(drbd_dev*,mdev);				\
-	if (atomic_sub_and_test(n, &mdev->unacked_cnt))		\
-		drbd_push_msock(mdev);				\
-	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt);
-#else
-#define dec_unacked(mdev)					\
+#define dec_unacked(mdev) do {					\
 	typecheck(drbd_dev*,mdev);				\
 	atomic_dec(&mdev->unacked_cnt);				\
-	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt)
+	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt); } while (0)
 
-#define sub_unacked(mdev, n)					\
+#define sub_unacked(mdev, n) do {				\
 	typecheck(drbd_dev*,mdev);				\
 	atomic_sub(n, &mdev->unacked_cnt);			\
-	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt)
-#endif
+	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt); } while (0)
 
 
 /**
