@@ -31,7 +31,7 @@
 
  */
 
-#include <linux/config.h>
+#include <linux/autoconf.h>
 #include <linux/module.h>
 #include <linux/version.h>
 
@@ -49,7 +49,9 @@
 #include <linux/mm_inline.h>
 #endif
 #include <linux/slab.h>
+#ifndef NO_MORE_DEV_FS
 #include <linux/devfs_fs_kernel.h>
+#endif
 
 #define __KERNEL_SYSCALLS__
 #include <linux/unistd.h>
@@ -144,9 +146,10 @@ int minor_count = 8;
 #endif
 int disable_bd_claim = 0;
 
+#ifndef NO_MORE_DEV_FS
 // devfs name
 char* drbd_devfs_name = "drbd";
-
+#endif
 
 // global panic flag
 volatile int drbd_did_panic = 0;
@@ -159,8 +162,8 @@ STATIC int *drbd_blocksizes;
 STATIC int *drbd_sizes;
 )
 struct Drbd_Conf *drbd_conf;
-kmem_cache_t *drbd_request_cache;
-kmem_cache_t *drbd_ee_cache;
+drbd_kmem_cache_t *drbd_request_cache;
+drbd_kmem_cache_t *drbd_ee_cache;
 mempool_t *drbd_request_mempool;
 
 STATIC struct block_device_operations drbd_ops = {
@@ -205,34 +208,19 @@ STATIC void tl_add(drbd_dev *mdev, drbd_request_t * new_item)
 {
 	struct drbd_barrier *b;
 
-	spin_lock_irq(&mdev->tl_lock);
+	spin_lock_irq(&mdev->req_lock);
 
 	b=mdev->newest_barrier;
 
 	new_item->barrier = b;
 	new_item->rq_status |= RQ_DRBD_IN_TL;
-	list_add(&new_item->w.list,&b->requests);
+	list_add_tail(&new_item->w.list,&b->requests);
 
 	if( b->n_req++ > mdev->conf.max_epoch_size ) {
 		set_bit(ISSUE_BARRIER,&mdev->flags);
 	}
 
-	spin_unlock_irq(&mdev->tl_lock);
-}
-
-STATIC void tl_cancel(drbd_dev *mdev, drbd_request_t * item)
-{
-	struct drbd_barrier *b;
-
-	spin_lock_irq(&mdev->tl_lock);
-
-	b=item->barrier;
-	b->n_req--;
-
-	list_del(&item->w.list);
-	item->rq_status &= ~RQ_DRBD_IN_TL;
-
-	spin_unlock_irq(&mdev->tl_lock);
+	spin_unlock_irq(&mdev->req_lock);
 }
 
 STATIC unsigned int tl_add_barrier(drbd_dev *mdev)
@@ -253,13 +241,13 @@ STATIC unsigned int tl_add_barrier(drbd_dev *mdev)
 	b->br_number=barrier_nr_issue;
 	b->n_req=0;
 
-	spin_lock_irq(&mdev->tl_lock);
+	spin_lock_irq(&mdev->req_lock);
 
 	bnr = mdev->newest_barrier->br_number;
 	mdev->newest_barrier->next = b;
 	mdev->newest_barrier = b;
 
-	spin_unlock_irq(&mdev->tl_lock);
+	spin_unlock_irq(&mdev->req_lock);
 
 	return bnr;
 }
@@ -268,48 +256,75 @@ void tl_release(drbd_dev *mdev,unsigned int barrier_nr,
 		       unsigned int set_size)
 {
 	struct drbd_barrier *b;
+	struct list_head *le,*tle;
+	drbd_request_t *req;
 
-	spin_lock_irq(&mdev->tl_lock);
+	spin_lock_irq(&mdev->req_lock);
 
 	b = mdev->oldest_barrier;
 	mdev->oldest_barrier = b->next;
 
-	list_del(&b->requests);
-	/* There could be requests on the list waiting for completion
-	   of the write to the local disk, to avoid corruptions of
-	   slab's data structures we have to remove the lists head */
-
-	spin_unlock_irq(&mdev->tl_lock);
+	list_for_each_safe(le,tle,&b->requests) {
+		req = list_entry(le, struct drbd_request,w.list);
+		/* master_bio already completed, but protocol != C,
+		 * so we had it still in the tl. need to list_del
+		 * anyways, there may be local io pending! */
+		list_del(&req->w.list);
+		req->rq_status &= ~RQ_DRBD_IN_TL;
+		if( (req->rq_status & RQ_DRBD_DONE) == RQ_DRBD_DONE ) {
+			D_ASSERT(req->master_bio == NULL);
+			INVALIDATE_MAGIC(req);
+			mempool_free(req,drbd_request_mempool);
+		}
+	}
+	spin_unlock_irq(&mdev->req_lock);
 
 	D_ASSERT(b->br_number == barrier_nr);
 	D_ASSERT(b->n_req == set_size);
 
+#if 1
+        if(b->br_number != barrier_nr) {
+                DUMPI(b->br_number);
+                DUMPI(barrier_nr);
+        }
+        if(b->n_req != set_size) {
+                DUMPI(b->n_req);
+                DUMPI(set_size);
+        }
+#endif
+
+	list_del(&b->requests);
 	kfree(b);
 }
 
-/* tl_dependence reports if this sector was present in the current
-   epoch.
-   As side effect it clears also the pointer to the request if it
-   was present in the transfert log. (Since tl_dependence indicates
-   that IO is complete and that drbd_end_req() should not be called
-   in case tl_clear has to be called due to interruption of the
-   communication)
-*/
-/* bool */
-int tl_dependence(drbd_dev *mdev, drbd_request_t * item)
+/* Since IO is frozen, nobody may modifiy the Transfer right log now...
+ */
+void tl_resend(drbd_dev *mdev)
 {
-	unsigned long flags;
-	int r=TRUE;
+	struct drbd_request *req;
+	struct list_head *le;
+	struct drbd_barrier *b;
+	int n_req;
 
-	spin_lock_irqsave(&mdev->tl_lock,flags);
+	D_ASSERT(test_bit(IO_FROZEN,&mdev->flags));
 
-	r = ( item->barrier == mdev->newest_barrier );
-	list_del(&item->w.list);
-
-	spin_unlock_irqrestore(&mdev->tl_lock,flags);
-	return r;
+	b = mdev->oldest_barrier;
+	while(1) {
+		n_req=0;
+		list_for_each(le, &b->requests) {
+			req = list_entry(le, struct drbd_request,w.list);
+			drbd_resend_dblock(mdev,req);
+			n_req++;
+		}
+		b->n_req = n_req; // only necessary for the oldest...
+		if( b == mdev->newest_barrier ) break;
+		drbd_resend_barrier(mdev,b);
+		b = b->next;
+	}
+	D_ASSERT(test_bit(IO_FROZEN,&mdev->flags));
 }
 
+
 void tl_clear(drbd_dev *mdev)
 {
 	struct list_head *le,*tle;
@@ -331,32 +346,31 @@ void tl_clear(drbd_dev *mdev)
 	new_first->br_number=4711;
 	new_first->n_req=0;
 
-	spin_lock_irq(&mdev->tl_lock);
+	spin_lock_irq(&mdev->req_lock);
 
 	b=mdev->oldest_barrier;
 	mdev->oldest_barrier = new_first;
 	mdev->newest_barrier = new_first;
 
-	spin_unlock_irq(&mdev->tl_lock);
-
 	inc_ap_pending(mdev); // Since we count the old first as well...
-
 	while ( b ) {
 		list_for_each_safe(le, tle, &b->requests) {
 			r = list_entry(le, struct drbd_request,w.list);
-			// bi_size and bi_sector are modified in bio_endio!
-			sector = drbd_req_get_sector(r);
-			size   = drbd_req_get_size(r);
+			// r may be freed within drbd_end_req
+			sector = r->sector;
+			size   = r->size;
+
+			r->rq_status &= ~RQ_DRBD_IN_TL;
+			list_del(&r->w.list);
+
 			if( !(r->rq_status & RQ_DRBD_SENT) ) {
-				if(mdev->conf.wire_protocol != DRBD_PROT_A )
-					dec_ap_pending(mdev);
-				drbd_end_req(r,RQ_DRBD_SENT,ERF_NOTLD|1, sector);
-				goto mark;
-			}
-			if(mdev->conf.wire_protocol != DRBD_PROT_C ) {
-			mark:
-				drbd_set_out_of_sync(mdev, sector, size);
-			}
+				_drbd_end_req(r,RQ_DRBD_SENT,1);
+			} else if ((r->rq_status & RQ_DRBD_DONE) == RQ_DRBD_DONE) {
+				D_ASSERT(r->master_bio == NULL);
+				INVALIDATE_MAGIC(r);
+				mempool_free(r,drbd_request_mempool);
+			} /* else: local io still pending */
+			drbd_set_out_of_sync(mdev, sector, size);
 		}
 		f=b;
 		b=b->next;
@@ -364,6 +378,7 @@ void tl_clear(drbd_dev *mdev)
 		kfree(f);
 		dec_ap_pending(mdev); // for the barrier
 	}
+	spin_unlock_irq(&mdev->req_lock);
 }
 
 /**
@@ -479,6 +494,19 @@ void _set_cstate(drbd_dev* mdev,Drbd_CState ns)
 // FIXME EXPLAIN
 		clear_bit(MD_IO_ALLOWED,&mdev->flags);
 	}
+
+	if(mdev->conf.on_disconnect == FreezeIO && mdev->state == Primary) {
+		if(os >= Connected && ns < Connected) {
+			set_bit(IO_FROZEN, &mdev->flags);
+		}
+	}
+
+	if( ns <= StandAlone && test_bit(IO_FROZEN, &mdev->flags)) {
+		WARN("Going to thaw IO, setting out of sync %d requests.\n",
+		     atomic_read(&mdev->ap_pending_cnt));
+		tl_clear(mdev);
+		clear_bit(IO_FROZEN, &mdev->flags);
+	}
 }
 
 STATIC int drbd_thread_setup(void* arg)
@@ -516,7 +544,6 @@ STATIC void drbd_thread_init(drbd_dev *mdev, struct Drbd_thread *thi,
 	thi->t_lock  = SPIN_LOCK_UNLOCKED;
 	thi->task    = NULL;
 	thi->t_state = None;
-	init_completion(&thi->startstop);
 
 	thi->function = func;
 	thi->mdev = mdev;
@@ -537,6 +564,7 @@ void drbd_thread_start(struct Drbd_thread *thi)
 	     thi->t_state); */
 
 	if (thi->t_state == None) {
+		init_completion(&thi->startstop);
 		D_ASSERT(thi->task == NULL);
 		thi->t_state = Running;
 		spin_unlock(&thi->t_lock);
@@ -562,10 +590,15 @@ void _drbd_thread_stop(struct Drbd_thread *thi, int restart,int wait)
 
 	spin_lock(&thi->t_lock);
 
-	/* INFO("%s [%d]: %s %d -> %d; %d\n",
+	/*
+	INFO("%s [%d]: %s[%s] %d -> %d; wait:%d; x->done:%d\n",
 	     current->comm, current->pid,
-	     thi->task ? thi->task->comm : "NULL", thi->t_state, ns, wait); */
-
+	     thi == &mdev->worker ? "worker" :
+	     thi == &mdev->asender ? "asender" :
+	     thi == &mdev->receiver ? "receiver" : "???",
+	     thi->task ? thi->task->comm : "NULL", thi->t_state, ns, wait,
+	     thi->startstop.done);
+	*/
 
 	if (thi->t_state == None) {
 		spin_unlock(&thi->t_lock);
@@ -586,10 +619,10 @@ void _drbd_thread_stop(struct Drbd_thread *thi, int restart,int wait)
 
 		thi->t_state = ns;
 		smp_mb();
-		if (thi->task != current)
+		if (thi->task != current) {
+			if(wait) init_completion(&thi->startstop);
 			force_sig(DRBD_SIGKILL,thi->task);
-		else
-			D_ASSERT(!wait);
+		} else D_ASSERT(!wait);
 
 	}
 	spin_unlock(&thi->t_lock);
@@ -813,14 +846,23 @@ int _drbd_send_barrier(drbd_dev *mdev)
 	int ok;
 	Drbd_Barrier_Packet p;
 
-	/* printk(KERN_DEBUG DEVICE_NAME": issuing a barrier\n"); */
 	/* tl_add_barrier() must be called with the sock_mutex aquired */
 	p.barrier=tl_add_barrier(mdev);
 
 	inc_ap_pending(mdev);
 	ok = _drbd_send_cmd(mdev,mdev->data.socket,Barrier,(Drbd_Header*)&p,sizeof(p),0);
 
-//	if (!ok) dec_ap_pending(mdev); // is done in tl_clear()
+	return ok;
+}
+
+int drbd_resend_barrier(drbd_dev *mdev,struct drbd_barrier *b)
+{
+	int ok;
+	Drbd_Barrier_Packet p;
+
+	p.barrier=b->br_number;
+	ok=drbd_send_cmd(mdev,USE_DATA_SOCKET, Barrier,(Drbd_Header*)&p,sizeof(p));
+
 	return ok;
 }
 
@@ -1039,10 +1081,9 @@ int drbd_send_dblock(drbd_dev *mdev, drbd_request_t *req)
 
 	p.head.magic   = BE_DRBD_MAGIC;
 	p.head.command = cpu_to_be16(Data);
-	p.head.length  = cpu_to_be16( sizeof(p)-sizeof(Drbd_Header)
-				     + drbd_req_get_size(req) );
+	p.head.length  = cpu_to_be16(sizeof(p)-sizeof(Drbd_Header)+req->size);
 
-	p.sector   = cpu_to_be64(drbd_req_get_sector(req));
+	p.sector   = cpu_to_be64(req->sector);
 	p.block_id = (unsigned long)req;
 
 	/* About tl_add():
@@ -1054,8 +1095,7 @@ int drbd_send_dblock(drbd_dev *mdev, drbd_request_t *req)
 	   tl_ datastructure (=> We would want to remove it before it
 	   is there!)
 	3. Q: Why can we add it to tl_ even when drbd_send() might fail ?
-	      There could be a tl_cancel() to remove it within the semaphore!
-	   A: If drbd_send fails, we will loose the connection. Then
+	   A: If drbd_send fails, we will lose the connection. Then
 	      tl_cear() will simulate a RQ_DRBD_SEND and set it out of sync
 	      for everything in the data structure.
 	*/
@@ -1069,42 +1109,107 @@ int drbd_send_dblock(drbd_dev *mdev, drbd_request_t *req)
 
 	old_blocked = drbd_block_all_signals();
 	down(&mdev->data.mutex);
+
+	/* drbd_disconnect may have freed that socket while we were waiting
+	 * in down(). we have to check that, to avoid a race with tl_clear
+	 * cleaning up before we can tl_add */
+	if (unlikely(!mdev->data.socket)) {
+                /* this req is not in the tl, tl_clear cannot find it.
+                 *
+                 * For Freeze-IO,
+                 * we add the req to the tl anyways, possibly creating
+                 * a new epoch, and then do nothing.  It will then still be
+                 * submitted locally, but it cannot possibly signal
+                 * completion to the upper layers before it was
+                 * resent (tl_resend) or canceled (tl_clear).
+                 *
+                 * Otherwise,
+                 * we cannot just tl_add it here, either, because tl_clear
+                 * might be done already.  so we have to mark this request
+                 * "SENT" here, otherwise it won't ever complete.
+                 * FIXME if we are Diskless, we complete a WRITE
+                 * as successful here, that has never been written! */
+		if (test_bit(IO_FROZEN,&mdev->flags)) {
+			if(test_and_clear_bit(ISSUE_BARRIER,&mdev->flags)) {
+				inc_ap_pending(mdev);
+				tl_add_barrier(mdev);
+			}
+			tl_add(mdev,req);
+		} else {
+			drbd_set_out_of_sync(mdev, req->sector, req->size);
+			drbd_end_req(req,RQ_DRBD_SENT,1);
+		}
+		goto out;
+	}
+
 	spin_lock(&mdev->send_task_lock);
 	mdev->send_task=current;
 	spin_unlock(&mdev->send_task_lock);
 
 	if(test_and_clear_bit(ISSUE_BARRIER,&mdev->flags))
 		ok = _drbd_send_barrier(mdev);
+
+	/* unconditional tl_add! */
+	tl_add(mdev,req);
 	if(ok) {
-		tl_add(mdev,req);
 		dump_packet(mdev,mdev->data.socket,0,(void*)&p, __FILE__, __LINE__);
 		set_bit(UNPLUG_REMOTE,&mdev->flags);
 		ok = sizeof(p) == drbd_send(mdev,mdev->data.socket,&p,sizeof(p),MSG_MORE);
 		if(ok) {
 			if(mdev->conf.wire_protocol == DRBD_PROT_A) {
 				ok = _drbd_send_bio(mdev,drbd_req_private_bio(req));
+				/* drbd_end_req has to be within the data.mutex,
+				 * which protects us from a concurrently running tl_clear */
+				if (ok) drbd_end_req(req,RQ_DRBD_SENT,1);
 			} else {
 				ok = _drbd_send_zc_bio(mdev,drbd_req_private_bio(req));
 			}
 		}
-		if(!ok) tl_cancel(mdev,req);
 	}
-	if (!ok) {
-		drbd_set_out_of_sync(mdev,
-				     drbd_req_get_sector(req),
-				     drbd_req_get_size(req));
-		drbd_end_req(req,RQ_DRBD_SENT,ERF_NOTLD|1,
-			     drbd_req_get_sector(req));
+	/* if (!ok) ... cleanup done in tl_clear */
+	spin_lock(&mdev->send_task_lock);
+	mdev->send_task=NULL;
+	spin_unlock(&mdev->send_task_lock);
+
+  out:
+	up(&mdev->data.mutex);
+	restore_old_sigset(old_blocked);
+	return ok;
+}
+
+int drbd_resend_dblock(drbd_dev *mdev, drbd_request_t *req)
+{
+	int ok;
+	Drbd_Data_Packet p;
+
+	p.head.magic   = BE_DRBD_MAGIC;
+	p.head.command = cpu_to_be16(Data);
+	p.head.length  = cpu_to_be16(sizeof(p)-sizeof(Drbd_Header)+req->size);
+
+	p.sector   = cpu_to_be64(req->sector);
+	p.block_id = (unsigned long)req;
+
+	down(&mdev->data.mutex);
+
+	spin_lock(&mdev->send_task_lock);
+	mdev->send_task=current;
+	spin_unlock(&mdev->send_task_lock);
+
+	dump_packet(mdev,mdev->data.socket,0,(void*)&p, __FILE__, __LINE__);
+	ok = sizeof(p) == drbd_send(mdev,mdev->data.socket,&p,sizeof(p),MSG_MORE);
+	if(ok) {
+		ok = _drbd_send_bio(mdev,req->master_bio);
 	}
+
 	spin_lock(&mdev->send_task_lock);
 	mdev->send_task=NULL;
 	spin_unlock(&mdev->send_task_lock);
 
 	up(&mdev->data.mutex);
-	restore_old_sigset(old_blocked);
 	return ok;
 }
 
+
 int drbd_send_block(drbd_dev *mdev, Drbd_Packet_Cmd cmd,
 		    struct Tl_epoch_entry *e)
 {
@@ -1375,7 +1480,6 @@ void drbd_init_set_defaults(drbd_dev *mdev)
 	sema_init(&mdev->meta.work.s,0);
 
 	mdev->al_lock        = SPIN_LOCK_UNLOCKED;
-	mdev->tl_lock        = SPIN_LOCK_UNLOCKED;
 	mdev->ee_lock        = SPIN_LOCK_UNLOCKED;
 	mdev->req_lock       = SPIN_LOCK_UNLOCKED;
 	mdev->pr_lock        = SPIN_LOCK_UNLOCKED;
@@ -1523,13 +1627,8 @@ void drbd_destroy_mempools(void)
 {
 	if (drbd_request_mempool)
 		mempool_destroy(drbd_request_mempool);
-	if (drbd_ee_cache && kmem_cache_destroy(drbd_ee_cache))
-		printk(KERN_ERR DEVICE_NAME
-		       ": kmem_cache_destroy(drbd_ee_cache) FAILED\n");
-	if (drbd_request_cache && kmem_cache_destroy(drbd_request_cache))
-		printk(KERN_ERR DEVICE_NAME
-		       ": kmem_cache_destroy(drbd_request_cache) FAILED\n");
-	// FIXME what can we do if we fail to destroy them?
+	if (drbd_ee_cache) kmem_cache_destroy(drbd_ee_cache);
+	if (drbd_request_cache) kmem_cache_destroy(drbd_request_cache);
 
 	drbd_request_mempool = NULL;
 	drbd_ee_cache        = NULL;
@@ -1695,7 +1794,9 @@ NOT_IN_26(
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 	devfs_unregister(devfs_handle);
 #else
+#ifndef NO_MORE_DEV_FS
 	devfs_remove(drbd_devfs_name);
+#endif
 #endif
 
 	if (unregister_blkdev(MAJOR_NR, DEVICE_NAME) != 0)
@@ -1769,7 +1870,9 @@ int __init drbd_init(void)
 		return err;
 	}
 
+#ifndef NO_MORE_DEV_FS
 	drbd_devfs_name = (major_nr == NBD_MAJOR) ? "nbd" : "drbd";
+#endif
 
 	/*
 	 * allocate all necessary structs
@@ -1791,7 +1894,9 @@ int __init drbd_init(void)
 	if (unlikely(!drbd_blocksizes)) goto Enomem;
 #else
 
+#ifndef NO_MORE_DEV_FS
 	devfs_mk_dir(drbd_devfs_name);
+#endif
 
 	for (i = 0; i < minor_count; i++) {
 		drbd_dev    *mdev = drbd_conf + i;
@@ -1814,7 +1919,9 @@ int __init drbd_init(void)
 		disk->first_minor = i;
 		disk->fops = &drbd_ops;
 		sprintf(disk->disk_name, DEVICE_NAME "%d", i);
+#ifndef NO_MORE_DEV_FS
 		sprintf(disk->devfs_name, "%s/%d", drbd_devfs_name, i);
+#endif
 		disk->private_data = mdev;
 		add_disk(disk);
 
