@@ -25,7 +25,7 @@
  */
 
 
-#include <linux/config.h>
+#include <linux/autoconf.h>
 #include <linux/module.h>
 
 #include <asm/uaccess.h>
@@ -810,7 +810,7 @@ int drbd_connect(drbd_dev *mdev)
 
 	drbd_thread_start(&mdev->asender);
 
-	drbd_send_param(mdev,0);
+	drbd_send_param(mdev,test_bit(IO_FROZEN, &mdev->flags) ? 4 : 0);
 	clear_bit(USE_DEGR_WFC_T,&mdev->flags);
 
 	return 1;
@@ -924,7 +924,7 @@ STATIC int recv_dless_read(drbd_dev *mdev, drbd_request_t *req,
 
 	bio = req->master_bio;
 
-	D_ASSERT( sector == drbd_req_get_sector(req) );
+	D_ASSERT( sector == req->sector );
 
 	rr=drbd_recv(mdev,drbd_bio_kmap(bio),data_size);
 	drbd_bio_kunmap(bio);
@@ -1203,7 +1203,6 @@ STATIC int receive_DataRequest(drbd_dev *mdev,Drbd_Header *h)
 		return FALSE;
 	}
 	e->block_id = p->block_id; // no meaning on this side, pr* on partner
-	list_add(&e->w.list,&mdev->read_ee);
 	spin_unlock_irq(&mdev->ee_lock);
 
 	if(!inc_local(mdev) || (mdev->gen_cnt[Flags] & MDF_Consistent) == 0) {
@@ -1245,6 +1244,9 @@ STATIC int receive_DataRequest(drbd_dev *mdev,Drbd_Header *h)
 
 	mdev->read_cnt += size >> 9;
 	inc_unacked(mdev);
+	spin_lock_irq(&mdev->ee_lock);
+	list_add(&e->w.list,&mdev->read_ee);
+	spin_unlock_irq(&mdev->ee_lock);
 	drbd_generic_make_request(READ,&e->private_bio);
 	if (atomic_read(&mdev->local_cnt) >= (mdev->conf.max_epoch_size>>4) ) {
 		drbd_kick_lo(mdev);
@@ -1526,6 +1528,8 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 	}
 	if(be32_to_cpu(p->flags)&2) consider_sync=1;
 
+	if(be32_to_cpu(p->flags)&4) consider_sync=0;
+
 	// XXX harmless race with ioctl ...
 	mdev->sync_conf.rate  =
 		max_t(int,mdev->sync_conf.rate, be32_to_cpu(p->sync_rate));
@@ -1581,13 +1585,30 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 
 	if (mdev->cstate == WFReportParams) {
 		INFO("Connection established.\n");
+
+		if(test_bit(IO_FROZEN, &mdev->flags)) {
+			WARN("Going to thaw IO, resending %d requests.\n",
+			     atomic_read(&mdev->ap_pending_cnt));
+			tl_resend(mdev);
+			set_bit(ISSUE_BARRIER,&mdev->flags);
+			if (mdev->cstate == WFReportParams) {
+				consider_sync = 0;
+			} else return FALSE;
+		}
 	}
 
 	if (consider_sync) {
 		if (!drbd_sync_handshake(mdev,p)) return FALSE;
 	}
 
-	if (mdev->cstate == WFReportParams) set_cstate(mdev,Connected);
+	if (mdev->cstate == WFReportParams) {
+		set_cstate(mdev,Connected);
+		if(test_bit(IO_FROZEN, &mdev->flags)) {
+			clear_bit(IO_FROZEN, &mdev->flags);
+			drbd_thaw_frozen_reqs(mdev);
+			consider_sync = 0;
+		}
+	}
 
 	oo_state = mdev->o_state;
 	mdev->o_state = be32_to_cpu(p->state);
@@ -1867,11 +1888,20 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	drbd_wait_ee(mdev,&mdev->sync_ee);
 	drbd_clear_done_ee(mdev);
 
-	// primary
-	tl_clear(mdev);
-	clear_bit(ISSUE_BARRIER,&mdev->flags);
-	wait_event( mdev->cstate_wait, atomic_read(&mdev->ap_pending_cnt)==0 );
-	D_ASSERT(mdev->oldest_barrier->n_req == 0);
+	if(test_bit(IO_FROZEN, &mdev->flags)) {
+		WARN("IO frozen with ap_pending_cnt = %d\n",
+		     atomic_read(&mdev->ap_pending_cnt));
+	} else {
+		tl_clear(mdev);
+		clear_bit(ISSUE_BARRIER,&mdev->flags);
+		wait_event( mdev->cstate_wait, atomic_read(&mdev->ap_pending_cnt)==0 );
+		D_ASSERT(mdev->oldest_barrier->n_req == 0);
+
+		if(atomic_read(&mdev->ap_pending_cnt)) {
+			ERR("ap_pending_cnt = %d\n",atomic_read(&mdev->ap_pending_cnt));
+			atomic_set(&mdev->ap_pending_cnt,0);
+		}
+	}
 
 	// both
 	clear_bit(PARTNER_CONSISTENT, &mdev->flags);
@@ -1903,11 +1933,6 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	   on the fly. */
 	atomic_set(&mdev->rs_pending_cnt,0);
 
-	if(atomic_read(&mdev->ap_pending_cnt)) {
-		ERR("ap_pending_cnt = %d\n",atomic_read(&mdev->ap_pending_cnt));
-		atomic_set(&mdev->ap_pending_cnt,0);
-	}
-
 	wake_up(&mdev->cstate_wait);
 
 	if ( mdev->state == Primary &&
@@ -2152,7 +2177,13 @@ STATIC int got_BlockAck(drbd_dev *mdev, Drbd_Header* h)
 
 			ERR_IF (!VALID_POINTER(req)) return FALSE;
 
-			drbd_end_req(req, RQ_DRBD_SENT, 1, sector);
+			spin_lock_irq(&mdev->req_lock);
+			if (mdev->conf.wire_protocol == DRBD_PROT_C) {
+				req->rq_status &= ~RQ_DRBD_IN_TL;
+				list_del(&req->w.list);
+			}
+			_drbd_end_req(req, RQ_DRBD_SENT, 1);
+			spin_unlock_irq(&mdev->req_lock);
 
 			if (test_bit(SYNC_STARTED,&mdev->flags) &&
 			    mdev->conf.wire_protocol == DRBD_PROT_C)
@@ -2164,7 +2195,6 @@ STATIC int got_BlockAck(drbd_dev *mdev, Drbd_Header* h)
 		dec_rs_pending(mdev);
 	} else {
 		D_ASSERT(mdev->conf.wire_protocol != DRBD_PROT_A);
-		dec_ap_pending(mdev);
 	}
 	return TRUE;
 }
@@ -2313,12 +2343,12 @@ int drbd_asender(struct Drbd_thread *thi)
 		 * this needs to be fixed properly, I'd vote for a separate
 		 * msock sender thread, but others will frown upon yet an other
 		 * kernel thread...
-		 *	-- lge
+		 *      -- lge
 		 */
 		set_bit(SIGNAL_ASENDER, &mdev->flags);
-
+		
 		if (!drbd_process_ee(mdev,0)) goto err;
-
+		
 		rv = drbd_recv_short(mdev,buf,expect-received);
 		clear_bit(SIGNAL_ASENDER, &mdev->flags);
 
